"""CLI module for calculating beads volumes"""
import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

SAMPLE_VOLUME_BOUNDARY = 50.0
ELUTION_VOLUME = 40.0
ELUTION_VOLUME_FACTOR = 0.8


def calculate_elution_volume(sample_volume: float) -> float:
    """Calculates the elution volume based on the sample volume"""

    return (
        ELUTION_VOLUME
        if sample_volume < SAMPLE_VOLUME_BOUNDARY
        else ELUTION_VOLUME_FACTOR * sample_volume
    )


def calculate_water_volume(sample_volume: float) -> float:
    """Calculates the H20 volume based on the sample volume"""
    return (
        SAMPLE_VOLUME_BOUNDARY - sample_volume
        if sample_volume < SAMPLE_VOLUME_BOUNDARY
        else 0.0
    )


def calculate_beads_volume(sample_volume: float, h2o_volume: float) -> float:
    """Calculates the bead volume bases on sample volume and H2O volume"""
    return 2 * (sample_volume + h2o_volume)


def calculate_volumes(artifacts: List[Artifact]):
    """Calculates beads volume, water volume and elution volume"""

    missing_udfs = 0

    for artifact in artifacts:
        sample_volume: float = artifact.udf.get("Sample Volume (ul)")
        if sample_volume is None:
            missing_udfs += 1
            continue
        h2o_volume = calculate_water_volume(sample_volume)
        elution_volume = calculate_elution_volume(sample_volume)
        beads_volume = calculate_beads_volume(sample_volume, h2o_volume)
        artifact.udf["Volume Elution (ul)"] = elution_volume
        artifact.udf["Volume H2O (ul)"] = h2o_volume
        artifact.udf["Volume Beads (ul)"] = beads_volume
        artifact.put()

    if missing_udfs:
        raise MissingUDFsError(
            f'Udf "Sample Volume (ul)" missing for {missing_udfs} out of {len(artifacts)} samples'
        )


@click.command()
@click.pass_context
def calculate_beads(context: click.Context):
    """Bead volume calculations for the step for Buffer Exchange TWIST"""

    LOG.info(f"Running {context.command_path} with params: {context.params}")

    process = context.obj["process"]

    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_volumes(artifacts=artifacts)
        message = "Beads volumes have been calculated."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from typing import Optional

from genologics.entities import Artifact, Process
from pydantic import Field, validator

from cg_lims.models.api.master_steps.base_step import BaseStep, get_process_udf
from cg_lims.get.artifacts import get_latest_artifact


class CaptureandWashTWIST(BaseStep):
    "Capture and Wash TWIST v2"
    artifact: Optional[Artifact]
    process: Optional[Process]
    enrichment_kit: Optional[str] = Field(None, alias="Twist enrichment kit")
    hybridization_time: Optional[str] = Field(None, alias="Total hybridization time (h)")

    @validator("artifact", always=True)
    def get_artifact(cls, v, values):
        try:
            return get_latest_artifact(
                lims=values.get("lims"),
                sample_id=values.get("sample_id"),
                process_types=["Capture and Wash TWIST v2"],
            )
        except:
            return None

    @validator("process", always=True)
    def get_process(cls, v, values):
        if values.get("artifact"):
            return values.get("artifact").parent_process
        return None

    @validator("enrichment_kit", always=True)
    def get_enrichment_kit(cls, v, values):
        return get_process_udf(values.get("process"), "Twist enrichment kit")

    @validator("hybridization_time", always=True)
    def get_hybridization_time(cls, v, values):
        return get_process_udf(values.get("process"), "Total hybridization time (h)")



import datetime as dt
import logging
from typing import Optional, Tuple

from genologics.entities import Artifact, Sample
from requests.exceptions import HTTPError

LOG = logging.getLogger(__name__)


def get_udf_value(sample: Sample, get_string: str):
    """Get sample udf based on udf name"""
    try:
        return sample.udf.get(get_string)
    except HTTPError:
        LOG.info("Sample %s doesn't have the udf %s" % (sample.id, get_string))
        return None


def get_prepared_date(sample: Sample) -> Optional[dt.date]:
    """Get the date when a sample was prepared in the lab."""

    return get_udf_value(sample, "Library Prep Finished")


def get_received_date(sample: Sample) -> Optional[dt.date]:
    """Get the date when a sample was received."""

    return get_udf_value(sample, "Received at")


def get_delivery_date(sample: Sample) -> Optional[dt.date]:
    """Get delivery date for a sample."""
    return get_udf_value(sample, "Delivered at")


def get_sequenced_date(sample: Sample) -> Optional[dt.date]:
    """Get the date when a sample was sequenced."""
    return get_udf_value(sample, "Sequencing Finished")


def get_sample_comment(sample: Sample) -> Optional[str]:
    """Get sample comment udf"""
    return get_udf_value(sample, "comment")


def get_processing_time(sample: Sample) -> Optional[dt.datetime]:
    """Get the time it takes to process a sample"""

    received_at = get_received_date(sample)
    delivery_date = get_delivery_date(sample)
    if received_at and delivery_date:
        return delivery_date - received_at
    LOG.info(
        "Could not get received date or delivery date to generate the processing time for sample %s."
        % sample.id
    )
    return None


def get_artifact_well(artifact: Artifact) -> str:
    """Parsing out the well position from LocationDescriptor"""

    location: Tuple = artifact.location
    return location[1].replace(":", "")


def get_index_well(artifact: Artifact):
    """Parsing out the index well position from the reagent label string which
    typically looks like this: 'A05 IDT_10nt_446 (AGCGTGACCT-CCATCCGAGT)'
    """

    if artifact.reagent_labels:
        # Assuming one reagent label per artifact (reagent_labels is a list):
        reagent_label = artifact.reagent_labels[0]

        # Getting the index well:
        index_well_with_zero = reagent_label.split(" ")[0]

        # Picking out column and removing zeros by int():
        index_well_col = int(index_well_with_zero[1:])
        index_well_row = index_well_with_zero[0]
        return f"{index_well_row}{index_well_col}"
    else:
        return "-"



from .buffer_exchange_v2 import BufferExchange
from .pool_samples_twist_v2 import PoolsamplesforhybridizationTWIST
from .bead_purification_twist_v2 import BeadPurificationTWIST
from .hybridize_library_twist_v2 import HybridizeLibraryTWIST
from .aliquot_samples_for_enzymatic_fragmentation_twist import (
    AliquotsamplesforenzymaticfragmentationTWIST,
)
from .capture_and_wash_twist_v2 import CaptureandWashTWIST
from .kapa_library_preparation_twist_v1 import KAPALibraryPreparation



from genologics.entities import Process
from genologics.lims import Lims
from tests.conftest import server

import pytest

from cg_lims.exceptions import MissingArtifactError
from cg_lims.get.artifacts import get_artifacts, get_latest_artifact


def test_get_latest_artifact(lims: Lims):
    # GIVEN a lims with a sample that has been run through the same
    # type of process three times but on different dates.
    server("test_get_artifacts")
    sample_id = "ACC7236A52"
    process_type = "CG002 - Sort HiSeq Samples"
    last_date = "2020-12-28"

    # WHEN running get_latest_artifact with the sample id and the process type name
    latest_artifact = get_latest_artifact(
        lims=lims, sample_id=sample_id, process_types=[process_type]
    )

    # THEN the artifact from the latest process will be returned
    assert latest_artifact.parent_process.date_run == last_date


def test_get_latest_artifact_no_artifacts(lims: Lims):
    # GIVEN a lims with no artifact related to a given sample id
    server("test_get_artifacts")
    sample_id = "SampleNotRelatedToArtifacts"
    process_type = "CG002 - Sort HiSeq Samples"

    # WHEN running get_latest_artifact
    # THEN MissingArtifactError is raised
    with pytest.raises(MissingArtifactError):
        latest_artifact = get_latest_artifact(
            lims=lims, sample_id=sample_id, process_types=process_type
        )


def test_get_artifacts_with_input_artifacts(lims: Lims):
    # GIVEN a process with one input artifacts
    server("test_get_artifacts")
    process = Process(lims, id="24-160122")

    # WHEN running get_artifacts
    input_artifacts = get_artifacts(process, input=True)

    # THEN assert input_artifacts are one
    assert len(input_artifacts) == 1


def test_get_artifacts_with_output_artifacts(lims: Lims):
    # GIVEN a process with five output artifacts
    server("test_get_artifacts")
    process = Process(lims, id="24-160122")

    # WHEN running get_artifacts
    output_artifacts = get_artifacts(process, input=False)

    # THEN assert output_artifacts are five
    assert len(output_artifacts) == 1



from genologics.lims import Lims
from pydantic import BaseModel


def get_artifact_udf(artifact, udf):
    if artifact:
        return artifact.udf.get(udf)
    return None


def get_process_udf(process, udf):
    if process:
        return process.udf.get(udf)
    return None


class BaseStep(BaseModel):
    sample_id: str
    lims: Lims

    class Config:
        arbitrary_types_allowed = True



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ArtifactUDFs(BaseModel):
    sample_concentration: Optional[float] = Field(None, alias="Concentration")


class ProcessUDFs(BaseModel):
    lot_nr_beads_buffer_exchange: Optional[str] = Field(None, alias="Lot Nr: Purification Beads")
    lot_nr_etoh_buffer_exchange: Optional[str] = Field(None, alias="Lot Nr: EtOH")
    lot_nr_h2o_buffer_exchange: Optional[str] = Field(None, alias="Lot Nr: Nuclease-free water")
    buffer_exchange_method: Optional[str] = Field(None, alias="Method document")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_buffer_exchange(lims: Lims, sample_id: str, prep_id: str) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Buffer Exchange v1",
        optional_step=True,
    )
    if not analyte.artifact:
        return None

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="buffer_exchange",
        workflow="Microbial"
    )



from tests.conftest import server
from pathlib import Path

from genologics.entities import Artifact
from cg_lims.EPPs.files.hamilton.normalization_file import get_file_data_and_write


def test_make_hamilton_normalization_file(hamilton_normalization_file, lims):
    # GIVEN: A file name. and a lims with two samples that has been run
    # through a process of type <amount_step>. There  they got udf: Amount needed (ng) set.
    # The two samples are now in another step where they have location and
    # reagent labels set.
    # (See entety_json_data for details)

    server("make_hamilton_normalization_file")
    file = Path("some_file_name")
    artifacts = [
        Artifact(lims, id="2-1450827"),
        Artifact(lims, id="2-1450824"),
        Artifact(lims, id="2-1450825"),
        Artifact(lims, id="2-1450826"),
    ]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(
        pool=False,
        destination_artifacts=artifacts,
        file=file,
        volume_udf="Sample Volume (ul)",
        buffer_udf="Volume H2O (ul)",
    )

    # THEN the file has been created and the file content is as expected.
    file_content = file.read_text()
    file.unlink()
    assert file_content == hamilton_normalization_file



from typing import List

from genologics.lims import Lims
from .library_preparation import get_library_prep_cov
from .pooling_and_cleanup import get_pooling_and_cleanup
from .reception_control import get_sample_artifact_fields
from ..base_step import BaseStep


def build_sars_cov_2_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a sars_cov_2 Prep."""
    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_pooling_and_cleanup(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_library_prep_cov(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
    ]
    return [document for document in step_documents if document is not None]



from typing import Optional

from genologics.lims import Lims
from pydantic import Field, BaseModel

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ProcessUDFs(BaseModel):
    performance_NA24143: Optional[str] = Field(None, alias="Batch no Prep Performance NA24143")
    GMCKsolid_HD827: Optional[str] = Field(None, alias="Batch no GMCKsolid-HD827")
    GMSlymphoid_HD829: Optional[str] = Field(None, alias="Batch no GMSlymphoid-HD829")
    GMSmyeloid_HD829: Optional[str] = Field(None, alias="Batch no GMSmyeloid-HD829")
    GMSsolid_HD832: Optional[str] = Field(None, alias="Batch no GMSsolid-HD832")
    aliquot_samples_library_preparation_method_2: str = Field(..., alias="Method document 2")
    aliquot_samples_library_preparation_method_1: str = Field(..., alias="Method document 1")
    lot_nr_h2o_aliquot_samples_fragmentation: str = Field(..., alias="Nuclease free water")


class ArtifactUDFs(BaseModel):
    amount_needed: Optional[float] = Field(None, alias="Amount needed (ng)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_aliquot_samples_for_enzymatic_fragmentation(
    lims: Lims, sample_id: str, prep_id: str
) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Aliquot samples for enzymatic fragmentation TWIST v2",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="aliquot_samples_for_enzymatic_fragmentation",
        workflow="TWIST",
    )



from genologics.entities import Process

from cg_lims.get.artifacts import get_artifacts
from tests.conftest import server
from pathlib import Path

from cg_lims.EPPs.files.hamilton.sars_cov2_prep_file import get_file_data_and_write


def test_make_hamilton_sars_cov2_pooling_file(hamilton_sars_cov2_pooling_file, lims):
    # GIVEN: A file name. and a lims with samples that has been run
    # through a pooling step with id 122-194650.

    server("hamilton_sars_cov2_pooling_step")
    file = Path("some_file_name")
    process = Process(lims, id="122-194650")
    artifacts = get_artifacts(process=process, input=False)

    # WHEN running get_file_data_and_write
    get_file_data_and_write(pool=True, file=file, destination_artifacts=artifacts)

    # THEN the file has been created and the file content is as expected.
    file_content = file.read_text()
    file.unlink()
    assert file_content == hamilton_sars_cov2_pooling_file


def test_make_hamilton_sars_cov2_indexing_file(hamilton_sars_cov2_indexing_file, lims):
    # GIVEN: A file name. and a lims with samples that has been run
    # through a indexing step with id 151-196204.

    server("hamilton_sars_cov2_indexing_step")
    file = Path("some_file_name")
    process = Process(lims, id="151-196204")
    artifacts = get_artifacts(process=process, input=False)

    # WHEN running get_file_data_and_write
    get_file_data_and_write(pool=False, file=file, destination_artifacts=artifacts)

    # THEN the file has been created and the file content is as expected.
    file_content = file.read_text()
    file.unlink()
    assert file_content == hamilton_sars_cov2_indexing_file



import pytest
from genologics.entities import Artifact

from cg_lims.EPPs.udf.calculate.calculate_water_volume_rna import (
    calculate_sample_and_water_volumes,
)
from cg_lims.exceptions import InvalidValueError, MissingUDFsError


def test_calculate_water_volume_rna_missing_concentration(artifact_1: Artifact):
    # GIVEN a list of artifacts with one artifact missing the udf 'Concentration'
    del artifact_1.udf["Concentration"]
    artifacts = [artifact_1]

    # WHEN calculating the sample and water volumes for all samples
    with pytest.raises(MissingUDFsError) as error_message:
        calculate_sample_and_water_volumes(artifacts)

    # THEN the MissingUDFsError exception should be raised
    assert (
        f"Could not apply calculation for 1 out of 1 sample(s): 'Concentration' is missing!"
        in error_message.value.message
    )


@pytest.mark.parametrize(
    "concentration",
    [
        100,
    ],
)
def test_calculate_water_volume_rna_amount_needed_missing(
    concentration, artifact_1: Artifact
):
    # GIVEN a list of artifacts with one artifact having no 'Amount Needed (ng)' udf
    artifact_1.udf["Concentration"] = concentration
    del artifact_1.udf["Amount needed (ng)"]
    artifacts = [artifact_1]

    # WHEN calculating sample and water volumes for all samples
    with pytest.raises(InvalidValueError) as error_message:
        calculate_sample_and_water_volumes(artifacts)

    # THEN the InvalidValueError exception should be raised
    assert (
        f"'Amount needed (ng)' missing or incorrect for one or more samples."
        in error_message.value.message
    )


@pytest.mark.parametrize(
    "concentration, amount_needed",
    [
        (100, 9001),
    ],
)
def test_calculate_water_volume_rna_(
    concentration, amount_needed, artifact_1: Artifact
):
    # GIVEN a list of artifacts with one artifact having an invalid 'Amount Needed (ng)' udf
    artifact_1.udf["Concentration"] = concentration
    artifact_1.udf["Amount needed (ng)"] = amount_needed
    artifacts = [artifact_1]

    # WHEN calculating the sample and water volumes for all samples
    with pytest.raises(InvalidValueError) as error_message:
        calculate_sample_and_water_volumes(artifacts)

    # THEN the InvalidValueError exception should be raised
    assert (
        f"'Amount needed (ng)' missing or incorrect for one or more samples."
        in error_message.value.message
    )


@pytest.mark.parametrize(
    "concentration, amount_needed, expected_water_volume, expected_sample_volume",
    [
        (100, 200, 23.0, 2.0),
        (100, 300, 22.0, 3.0),
        (100, 400, 21.0, 4.0),
        (100, 500, 20.0, 5.0),
    ],
)
def test_calculate_water_volume_rna(
    artifact_1: Artifact,
    concentration,
    amount_needed,
    expected_water_volume,
    expected_sample_volume,
):
    # GIVEN a list of artifacts with one artifact
    artifact_1.udf["Concentration"] = concentration
    artifact_1.udf["Amount needed (ng)"] = amount_needed
    artifacts = [artifact_1]

    # WHEN calculating the sample and water volumes for all samples
    calculate_sample_and_water_volumes(artifacts)

    # THEN the correct volumes should be set on the artifact
    assert artifact_1.udf["Volume H2O (ul)"] == expected_water_volume
    assert artifact_1.udf["Sample Volume (ul)"] == expected_sample_volume



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ProcessUDFs(BaseModel):
    """Aliquot Samples for Fragmentation (RNA) v1"""

    poly_a_capture: str = Field(..., alias="RGT no: Illumina PolyA capture")
    cdna_synthesis: str = Field(..., alias="RGT no: Illumina cDNA synthesis")
    water: str = Field(..., alias="Lot no: Nuclease free water")
    pcr_name: str = Field(..., alias="Thermal cycler name")
    ampure_beads: str = Field(..., alias="Lot no: AMPure XP-beads")
    et_oh: str = Field(..., alias="Lot no: EtOH")


class ArtifactUDFs(BaseModel):
    volume_h2o: float = Field(..., alias="Volume H2O (ul)")
    volume_sample: float = Field(..., alias="Sample Volume (ul)")


class ArnoldStep(
    BaseStep,
):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_aliquot_samples_for_enzymatic_fragmentation(
    lims: Lims, sample_id: str, prep_id: str
) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Aliquot Samples for Fragmentation (RNA) v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="aliquot_samples_for_fragmentation",
        workflow="RNA"
    )



from cg_lims.exceptions import FailingQCError, MissingUDFsError
from genologics.entities import Artifact, Sample
from genologics.lims import Lims
from cg_lims.EPPs.udf.calculate.twist_qc_amount import calculate_amount_and_set_qc
from tests.conftest import server

import pytest


@pytest.mark.parametrize("volume,concentration,source",
                         [(3, 1, "blood"), (10, 0.1, "cfDNA"), (50, 300, "blood"), (300, 2, "blood"),
                          (10, 10, "blood")])
def test_calculate_amount_and_set_qc_failing(volume: int, concentration: float, source: int, artifact_1: Artifact,
                                             sample_1: Sample):
    # GIVEN: A sample with udf Source: <source>,
    # and a related Artifact with udfs: Concentration: <concentration> and  Volume (ul): <volume>
    server('flat_tests')
    artifact_1.udf["Volume (ul)"] = volume
    artifact_1.udf["Concentration"] = concentration
    artifact_1.qc_flag == "UNKNOWN"
    artifact_1.put()

    sample_1.udf["Source"] = source
    sample_1.put()

    # THEN FailingQCError is being raised and the artifacts should be FAILED
    with pytest.raises(FailingQCError):
        calculate_amount_and_set_qc(artifacts=[artifact_1])
    assert artifact_1.qc_flag == "FAILED"


@pytest.mark.parametrize("volume,concentration,source", [(35, 10, "cfDNA"), (50, 50, "blood")])
def test_calculate_amount_and_set_qc_passing(volume: int, concentration: float, source: int, artifact_1: Artifact,
                                             sample_1: Sample):
    # GIVEN: A sample with udf Source: <source>,
    # and a related Artifact with udfs: Concentration: <concentration> and  Volume (ul): <volume>
    server('flat_tests')
    artifact_1.udf["Volume (ul)"] = volume
    artifact_1.udf["Concentration"] = concentration
    artifact_1.qc_flag == "UNKNOWN"
    artifact_1.put()

    sample_1.udf["Source"] = source
    sample_1.put()

    # WHEN running calculate_amount_and_set_qc
    calculate_amount_and_set_qc(artifacts=[artifact_1])

    # THEN the artifacts should be PASSED
    assert artifact_1.qc_flag == "PASSED"


def test_calculate_amount_and_set_qc_missing_udf(lims: Lims, artifact_1: Artifact, artifact_2: Artifact):
    # GIVEN: A list of two artifacts with missing udfs Volume (ul) or Concentration
    server('flat_tests')
    artifact_1.udf["Volume (ul)"] = 35
    artifact_1.put()
    artifact_2.udf["Concentration"] = 1
    artifact_2.put()

    artifacts = [artifact_1, artifact_2]

    # WHEN running calculate_amount_and_set_qc
    # THEN MissingUDFsError is being raised
    with pytest.raises(MissingUDFsError):
        calculate_amount_and_set_qc(artifacts=artifacts)



from typing import Optional

from genologics.lims import Lims
from pydantic import Field, BaseModel

from cg_lims.objects import BaseAnalyte

from cg_lims.models.arnold.prep.base_step import BaseStep


class SampleArtifactUDF(BaseModel):
    concentration: Optional[float] = Field(None, alias="Concentration")


class SampleArtifactFields(BaseStep):
    artifact_udfs: SampleArtifactUDF

    class Config:
        allow_population_by_field_name = True


def get_sample_artifact_fields(
    lims: Lims, sample_id: str, prep_id: str
) -> Optional[SampleArtifactFields]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        optional_step=True,
    )

    sample_artifact_udfs = SampleArtifactUDF(**analyte.artifact_udfs())
    if not sample_artifact_udfs.dict(exclude_none=True):
        return None

    return SampleArtifactFields(
        **analyte.base_fields(),
        artifact_udfs=sample_artifact_udfs,
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="reception_control",
        workflow="TWIST",
    )



from typing import Optional

from genologics.lims import Lims
from pydantic import Field, BaseModel

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ProcessUdfs(BaseModel):
    fragmentation_method: Optional[str] = Field(None, alias="Method document")
    fragmentation_time: Optional[float] = Field(None, alias="Fragmentation time (min)")
    fragmentation_kit: Optional[str] = Field(None, alias="KAPA HyperPlus Kit")
    fragmentation_instrument_hybridization: Optional[str] = Field(None, alias="Thermal cycler name")
    fragmentation_hamilton: Optional[str] = Field(None, alias="Hamilton")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUdfs

    class Config:
        allow_population_by_field_name = True


def get_enzymatic_fragmentation(lims: Lims, sample_id: str, prep_id: str) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Enzymatic fragmentation TWIST v2",
        optional_step=True,
    )
    if not analyte.artifact:
        return None

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUdfs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="enzymatic_fragmentation",
        workflow="TWIST",
    )



from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte

from cg_lims.models.arnold.prep.base_step import BaseStep


class SampleArtifactUDF(BaseModel):
    sample_concentration: float = Field(..., alias="Concentration")
    sample_amount: float = Field(..., alias="Amount (ng)")


class SampleArtifactFields(BaseStep):
    artifact_udfs: SampleArtifactUDF

    class Config:
        allow_population_by_field_name = True


def get_sample_artifact_fields(lims: Lims, sample_id: str, prep_id: str) -> SampleArtifactFields:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
    )

    return SampleArtifactFields(
        **analyte.base_fields(),
        artifact_udfs=SampleArtifactUDF(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="reception_control",
        workflow="WGS"
    )



from typing import List

from genologics.lims import Lims

from .fragment_dna_truseq_dna import get_fragemnt_dna_truseq
from .aliquot_sampels_for_covaris import get_aliquot_samples_for_covaris
from .reception_control import get_sample_artifact_fields
from .endrepair_size_selection_a_tailing_adapter_ligation import get_end_repair
from ..base_step import BaseStep


def build_wgs_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a sars_cov_2 Prep."""

    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_fragemnt_dna_truseq(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_aliquot_samples_for_covaris(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_end_repair(sample_id=sample_id, lims=lims, prep_id=prep_id),
    ]
    return [document for document in step_documents if document is not None]



"""Unit tests for cg_lims.EPPs.udf.calculate.calculate_microbial_aliquot_volumes"""

import pytest
from genologics.entities import Artifact

from cg_lims.EPPs.udf.calculate.calculate_microbial_aliquot_volumes import (
    calculate_volumes,
)
from cg_lims.exceptions import HighConcentrationError, MissingUDFsError

QC_FAILED = "FAILED"
QC_PASSED = "PASSED"


@pytest.mark.parametrize(
    "sample_concentration, sample_volume, total_volume, buffer_volume, qc_flag",
    [
        (0, 15, 15, 0, QC_PASSED),
        (1.999, 15, 15, 0, QC_PASSED),
        (2, 15, 15, 0, QC_PASSED),
        (2.001, 15, 15, 0, QC_PASSED),
        (2.30769, 15, 15, 0, QC_PASSED),
        (30 / 13, 15, 15, 0, QC_PASSED),
        (2.30770, 12.99995666681111, 15, 2.0000433331888896, QC_PASSED),
        (7.449, 4.027386226339106, 15, 10.972613773660894, QC_PASSED),
        (7.5, 4, 15, 11, QC_PASSED),
        (7.5001, 4, 15.0002, 11.0002, QC_PASSED),
        (59.999, 4, 119.998, 115.998, QC_PASSED),
        (60, 4, 120, 116, QC_PASSED),
    ],
)
def test_calculate_microbial_aliquot_volumes(
    sample_concentration: float,
    sample_volume: float,
    total_volume: float,
    buffer_volume: float,
    qc_flag: str,
    artifact_1: Artifact,
):
    # GIVEN a sample with a valid concentration
    artifact_1.udf["Concentration"] = sample_concentration
    artifacts = [artifact_1]

    # WHEN calculating the microbial aliquot volumes
    calculate_volumes(artifacts=artifacts)

    # THEN the sample udfs should be set to the correct values
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume
    assert artifact_1.udf.get("Volume Buffer (ul)") == buffer_volume
    assert artifact_1.udf.get("Total Volume (uL)") == total_volume
    assert artifact_1.qc_flag == qc_flag


@pytest.mark.parametrize(
    "sample_concentration",
    [
        None,
    ],
)
def test_calculate_microbial_aliquot_volumes_concentration_missing(
    sample_concentration: float, artifact_1: Artifact, caplog
):
    # GIVEN a sample with no concentration udf
    artifact_1.udf["Concentration"] = sample_concentration
    artifacts = [artifact_1]

    # WHEN calculating the microbial aliquot volumes
    with pytest.raises(MissingUDFsError) as error_message:
        calculate_volumes(artifacts=artifacts)

    # THEN a MissingUDFsError should be raised
    assert (
        "Could not apply calculations for 1 out of 1 sample(s): 'Concentration' is missing!"
        in error_message.value.message
    )


@pytest.mark.parametrize(
    "sample_concentration, qc_flag",
    [(60.001, QC_FAILED)],
)
def test_calculate_microbial_aliquot_volumes_concentration_too_high(
    sample_concentration: float,
    qc_flag: str,
    artifact_1: Artifact,
):
    # GIVEN a sample with a concentration value greater than the maximum allowed
    artifact_1.udf["Concentration"] = sample_concentration
    artifacts = [artifact_1]
    mu = "\u03BC"

    # WHEN calculating the microbial aliquot volumes
    with pytest.raises(HighConcentrationError) as error_message:
        calculate_volumes(artifacts=artifacts)

    # THEN a HighConcentrationError should be raised
    assert artifact_1.qc_flag == "FAILED"
    assert (
        f"Could not apply calculations for one or more sample(s): concentration too high (> "
        f"60 ng/{mu}l)! " in error_message.value.message
    )


@pytest.mark.parametrize(
    "sample_concentration, sample_volume, total_volume, buffer_volume, name, qc_flag",
    [
        (999, 0, 15, 15, "NTC-CG99", QC_PASSED),
    ],
)
def test_calculate_microbial_aliquot_volumes_ntc_sample(
    sample_concentration: float,
    sample_volume: float,
    total_volume: float,
    buffer_volume: float,
    name: str,
    qc_flag: str,
    artifact_1: Artifact,
):
    # GIVEN an NTC sample with a valid concentration
    artifact_1.udf["Concentration"] = sample_concentration
    artifact_1.samples[0].name = name
    artifacts = [artifact_1]

    # WHEN calculating the microbial aliquot volumes
    calculate_volumes(artifacts=artifacts)

    # THEN the sample udfs should be set to the correct values
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume
    assert artifact_1.udf.get("Volume Buffer (ul)") == buffer_volume
    assert artifact_1.udf.get("Total Volume (uL)") == total_volume
    assert artifact_1.qc_flag == qc_flag



"""
CLI module to calculate dilution volumes from Concentration udf

For an explanation of the sample volume calculation and setting of the QG flag, see
AM-document 1243 Method - Preparing Plate for Genotyping, section 3.3.2 Preparing the plate
"""

import logging
import sys
from typing import List, Optional

import click
from genologics.entities import Artifact
from pydantic import BaseModel, validator

from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

FINAL_CONCENTRATION = 4
MINIMUM_TOTAL_VOLUME = 15
QC_FAILED = "FAILED"
QC_PASSED = "PASSED"
STANDARD_SAMPLE_VOLUME = 3


class MafVolumes(BaseModel):
    """Pydantic model for calculating MAF volumes based on the sample concentration"""

    sample_concentration: float
    sample_volume: Optional[float]
    final_volume: Optional[float]
    water_volume: Optional[float]
    qc_flag: Optional[str]

    @validator("sample_concentration", always=True)
    def set_sample_concentration(cls, v, values):
        """Set sample concentration and handle low or missing concentration"""
        if v < FINAL_CONCENTRATION:
            message = "Too low or missing concentration"
            LOG.error(message)
            raise ValueError(message)
        return v

    @validator("sample_volume", always=True)
    def set_sample_volume(cls, v, values):
        """Calculate the sample volume. For an explanation about how sample volumes are
        determined for various sample concentration ranges, refer to AM document 1243 Method -
        Preparing Plate for Genotyping, section 3.3.2"""
        sample_concentration = values.get("sample_concentration")
        if sample_concentration < 20:
            return MINIMUM_TOTAL_VOLUME * FINAL_CONCENTRATION / sample_concentration
        elif 20 <= sample_concentration < 244:
            return 3
        elif 244 <= sample_concentration < 364:
            return 2
        elif 364 <= sample_concentration < 724:
            return 1
        elif 724 <= sample_concentration < 1444:
            return 0.5
        else:
            message = "Sample concentration not in valid range"
            LOG.error(message)
            raise ValueError(message)

    @validator("water_volume", always=True)
    def set_water_volume(cls, v, values):
        """Calculate the water volume for a sample"""
        return values.get("final_volume") - values.get("sample_volume")

    @validator("final_volume", always=True)
    def set_final_volume(cls, v, values):
        """Calculates the final volume for a sample"""
        return (
            values.get("sample_volume")
            * values.get("sample_concentration")
            / FINAL_CONCENTRATION
        )

    @validator("qc_flag", always=True)
    def set_qc_flag(cls, v, values):
        """Set the QC flag on a sample"""
        if values.get("sample_volume") == STANDARD_SAMPLE_VOLUME:
            return QC_PASSED
        else:
            return QC_FAILED


def calculate_volume(artifacts: List[Artifact]) -> None:
    """Determines the final volume, water volume, and sample volume"""
    failed_artifacts = []

    for artifact in artifacts:
        try:
            volumes = MafVolumes(sample_concentration=artifact.udf.get("Concentration"))
            artifact.qc_flag = volumes.qc_flag
            artifact.udf["Final Volume (uL)"] = volumes.final_volume
            artifact.udf["Volume H2O (ul)"] = volumes.water_volume
            artifact.udf["Volume of sample (ul)"] = volumes.sample_volume
            artifact.put()
        except Exception:
            LOG.warning(f"Could not calculate sample volume for sample {artifact.id}.")
            failed_artifacts.append(artifact)
            continue

    if failed_artifacts:
        raise LimsError(
            f"MAF volume calculations failed for {len(failed_artifacts)} samples, "
            f"{len(artifacts) - len(failed_artifacts)} updates successful. "
        )


@click.command()
@click.pass_context
def maf_calculate_volume(context: click.Context):
    """Calculate dilution volumes based on the Concentration udf"""
    LOG.info(f"Running {context.command_path} with params: {context.params}")
    process = context.obj["process"]
    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_volume(artifacts=artifacts)
        message = "MAF volumes have been calculated."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from typing import List

from genologics.lims import Lims

from .a_tailing_and_adapter_ligation import get_a_tailning_and_adapter_ligation
from .aliquot_samples_for_fragmentation import get_aliquot_samples_for_enzymatic_fragmentation
from .reception_control import get_sample_artifact_fields
from .normalization_of_samples_for_sequencing import get_normalization_of_samples
from ..base_step import BaseStep


def build_rna_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a RNA Prep."""
    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_normalization_of_samples(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_a_tailning_and_adapter_ligation(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_aliquot_samples_for_enzymatic_fragmentation(
            sample_id=sample_id, lims=lims, prep_id=prep_id
        ),
    ]
    return [document for document in step_documents if document is not None]



#!/usr/bin/env python

import logging
import sys
from typing import List

import click
from genologics.entities import Artifact, Sample
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingArtifactError
from cg_lims.get.artifacts import get_latest_artifact, get_sample_artifact
from cg_lims.get.samples import get_process_samples
from cg_lims.put.queue import queue_artifacts

LOG = logging.getLogger(__name__)


def get_pools_and_samples_to_queue(
    lims: Lims, process_type: List[str], samples: List[Sample]
) -> List[Artifact]:
    """Get samples and pools to place in sequence aggregation.
    Sort of specific script:
    Single arts                                       --> Next Step
    Cust001 - Pools --> split to pools from sort step --> Next Step
    Non RML - Pools --> split in uniq sample arts     --> Next Step

    Args:
        lims: Lims
        rerun_arts: List # Artifacts with Rerun flag
        process_type: List[str] # Name of step(s) before the requeue step
    """

    break_send_to_next_step = False
    send_to_next_step = []
    for sample in samples:
        pool_name = sample.udf.get("pool name")
        if pool_name:
            ## this is a RML - get pools from sort step
            try:
                artifact = get_latest_artifact(lims, sample.id, process_type)
            except MissingArtifactError as e:
                LOG.warning(e.message)
                break_send_to_next_step = True
                continue
        else:
            ## this is a pool (or a sample) and we want to pass its samples to next step
            artifact = get_sample_artifact(lims, sample)
        send_to_next_step.append(artifact)
    if break_send_to_next_step:
        raise MissingArtifactError("Issues getting pools and or samples to queue. See log")
    return list(set(send_to_next_step))


@click.command()
@options.workflow_id(help="Destination workflow id.")
@options.stage_id(help="Destination stage id.")
@options.process_types(
    help="The name(s) of the process type(s) from where we want to fetch the pools"
)
@click.pass_context
def place_samples_in_seq_agg(ctx, workflow_id, stage_id, process_types):
    """Queueing artifacts with given udf==True, to stage in workflow.
    Raising error if quiueing fails."""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    samples = get_process_samples(process)

    try:
        artifacts = get_pools_and_samples_to_queue(lims, process_types, samples)
        queue_artifacts(lims, artifacts, workflow_id, stage_id)
        LOG.info("Artifacts have been queued.")
        click.echo("Artifacts have been queued.")
    except LimsError as e:
        sys.exit(e.message)



from pathlib import Path
from typing import List

import click
from genologics.entities import Entity
from genologics.lims import Process
from pydantic import BaseModel


class ProcessFixure(BaseModel):
    samples: Path
    artifacts: Path
    containers: Path
    containertypes: Path
    processes: Path


def replace_str(file: Path, replace: str, replace_with: str) -> None:
    file.write_text(file.read_text().replace(replace, replace_with))


def build_file_structure(base_dir: str) -> ProcessFixure:
    processes = Path(f"{base_dir}/processes/")
    if not processes.exists():
        processes.mkdir(parents=True)
    artifacts = Path(f"{base_dir}/artifacts/")
    if not artifacts.exists():
        artifacts.mkdir(parents=True)
    containers = Path(f"{base_dir}/containers/")
    if not containers.exists():
        containers.mkdir(parents=True)
    containertypes = Path(f"{base_dir}/containertypes/")
    if not containertypes.exists():
        containertypes.mkdir(parents=True)
    samples = Path(f"{base_dir}/samples/")
    if not samples.exists():
        samples.mkdir(parents=True)
    return ProcessFixure(
        samples=samples,
        processes=processes,
        artifacts=artifacts,
        containers=containers,
        containertypes=containertypes,
    )


def add_file(entity: Entity, entity_dir: Path) -> None:
    new_file = Path(f"{entity_dir.absolute()}/{entity.id}.xml")
    with open(new_file.absolute(), "wb") as file:
        file.write(entity.xml())
    replace_str(
        file=new_file,
        replace="clinical-lims-stage.scilifelab.se",
        replace_with="127.0.0.1:8000",
    )
    replace_str(
        file=new_file,
        replace="https:",
        replace_with="http:",
    )


def add_entities(entities: List[Entity], entity_dir: Path):

    for entity in entities:
        entity.get()
        add_file(entity=entity, entity_dir=entity_dir)


@click.command()
@click.option("--process")
@click.option("--test_name")
@click.pass_context
def make_fixure(ctx, process: str, test_name: str):
    lims = ctx.obj["lims"]
    process = Process(lims=lims, id=process)
    process.get()
    fixture_dir: ProcessFixure = build_file_structure(base_dir=test_name)
    add_file(entity=process, entity_dir=fixture_dir.processes)
    artifacts = process.all_inputs() + process.all_outputs()
    add_entities(entities=artifacts, entity_dir=fixture_dir.artifacts)
    samples = []
    for artifact in artifacts:
        samples += artifact.samples
    add_entities(entities=samples, entity_dir=fixture_dir.samples)
    containers = {artifact.location[0] for artifact in artifacts}
    if None in containers:
        containers.remove(None)
    add_entities(entities=list(containers), entity_dir=fixture_dir.containers)
    container_types = list({container.type for container in containers})
    add_entities(entities=container_types, entity_dir=fixture_dir.containertypes)



from genologics.entities import Sample, Process

from cg_lims.EPPs.udf.set.set_sample_date import (
    set_prepared,
    set_sequenced,
    set_delivered,
)
from datetime import datetime as dt
from tests.conftest import (
    server,
)


def test_set_prepared(lims):
    # GIVEN: A lims with a sample: "ACC8454A1" with a no sequencing date and no delivery date.

    server("wgs_prep")
    sample = Sample(lims=lims, id="ACC8454A1")

    # WHEN running set_prepared
    set_prepared(sample)

    # THEN assert the prepared date was set to todays date
    assert sample.udf.get("Library Prep Finished") == dt.today().date()


def test_set_prepared_delivered_and_sequenced_were_set(lims):
    # GIVEN: A lims with a sample: "ACC8454A1" with a sequencing date and a delivery date.

    server("wgs_prep")
    sample = Sample(lims=lims, id="ACC8454A1")
    sample.udf["Delivered at"] = dt(2020, 6, 17).date()
    sample.udf["Sequencing Finished"] = dt(2020, 5, 17).date()
    sample.put()

    # WHEN running set_prepared
    set_prepared(sample)

    # THEN assert the prepared date was set to todays date but the other two dates were set to None!
    assert sample.udf.get("Library Prep Finished") == dt.today().date()
    assert sample.udf.get("Delivered at") is None
    assert sample.udf.get("Sequencing Finished") is None


def test_set_sequenced(lims):
    # GIVEN: A lims with a sample: "ACC8454A1" with a no sequencing date and no delivery date.

    server("wgs_prep")
    sample = Sample(lims=lims, id="ACC8454A1")

    # WHEN running set_sequenced
    set_sequenced(sample)

    # THEN assert the sequrncing date was set to todays date
    assert sample.udf.get("Sequencing Finished") == dt.today().date()


def test_set_sequenced_date_was_set(lims):
    # GIVEN: A lims with a sample: "ACC8454A1" with a delivery date.

    server("wgs_prep")
    old_date = dt(2020, 6, 17).date()
    sample = Sample(lims=lims, id="ACC8454A1")
    sample.udf["Sequencing Finished"] = old_date
    sample.put()

    # WHEN running set_sequenced
    set_sequenced(sample)

    # THEN assert the sequencing date was not updated!
    assert sample.udf.get("Sequencing Finished") != dt.today().date()
    assert sample.udf.get("Sequencing Finished") == old_date


def test_set_delivered(lims):
    # GIVEN: A lims with a sample: "ACC8454A1" and a process with a Date delivered udf set.

    server("wgs_prep")
    sample = Sample(lims=lims, id="ACC8454A1")
    process = Process(lims=lims, id="24-240272")
    process.udf["Date delivered"] = dt.today().date()
    process.put()

    # WHEN running set_delivered
    set_delivered(sample=sample, process=process)

    # THEN assert the delivery date was set on the sample
    assert sample.udf.get("Delivered at") == process.udf["Date delivered"]



import logging
from typing import List

import click
from genologics.lims import Process, Sample
import requests
from requests import Response
import json
from cg_lims.exceptions import LimsError
from cg_lims.get.samples import get_process_samples
from cg_lims.models.arnold.sample import ArnoldSample

LOG = logging.getLogger(__name__)


@click.command()
@click.pass_context
def sample(ctx):
    """Creating Sample documents in the arnold Sample collection."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process: Process = ctx.obj["process"]
    arnold_host: str = ctx.obj["arnold_host"]
    samples: List[Sample] = get_process_samples(process=process)

    sample_documents = []
    for sample in samples:
        arnold_sample = ArnoldSample(**dict(sample.udf.items()), sample_id=sample.id, id=sample.id)
        sample_documents.append(arnold_sample.dict(exclude_none=True))

    response: Response = requests.put(
        url=f"{arnold_host}/samples",
        headers={"Content-Type": "application/json"},
        data=json.dumps(sample_documents),
    )
    if not response.ok:
        LOG.info(response.text)
        raise LimsError(response.text)

    LOG.info("Arnold output: %s", response.text)
    click.echo("Sample documents inserted to arnold database")



from typing import Optional

from genologics.entities import Artifact, Process
from pydantic import Field, validator

from cg_lims.models.api.master_steps.base_step import (
    get_artifact_udf,
    BaseStep,
    get_process_udf,
)
from cg_lims.get.artifacts import get_latest_artifact


class AliquotsamplesforenzymaticfragmentationTWIST(BaseStep):
    "Aliquot samples for enzymatic fragmentation TWIST"
    artifact: Optional[Artifact]
    process: Optional[Process]
    performance_NA24143: Optional[str] = Field(None, alias="Batch no Prep Performance NA24143")
    GMCKsolid_HD827: Optional[str] = Field(None, alias="Batch no GMCKsolid-HD827")
    GMSlymphoid_HD829: Optional[str] = Field(None, alias="Batch no GMSlymphoid-HD829")
    GMSmyeloid_HD829: Optional[str] = Field(None, alias="Batch no GMSmyeloid-HD829")
    amount_needed: Optional[str] = Field(None, alias="Amount needed (ng)")

    @validator("artifact", always=True, pre=True)
    def get_artifact(cls, v, values):
        try:
            return get_latest_artifact(
                lims=values.get("lims"),
                sample_id=values.get("sample_id"),
                process_types=["Aliquot samples for enzymatic fragmentation TWIST v2"],
            )
        except:
            return None

    @validator("process", always=True)
    def get_process(cls, v, values):
        if values.get("artifact"):
            return values.get("artifact").parent_process
        return None

    @validator("performance_NA24143", always=True)
    def get_performance_NA24143(cls, v, values):
        return get_process_udf(values.get("process"), "Batch no Prep Performance NA24143")

    @validator("GMCKsolid_HD827", always=True)
    def get_GMCKsolid_HD827(cls, v, values):
        return get_process_udf(values.get("process"), "Batch no GMCKsolid-HD827")

    @validator("GMSlymphoid_HD829", always=True)
    def get_GMSlymphoid_HD829(cls, v, values):
        return get_process_udf(values.get("process"), "Batch no GMSlymphoid-HD829")

    @validator("GMSmyeloid_HD829", always=True)
    def get_GMSmyeloid_HD829(cls, v, values):
        return get_process_udf(values.get("process"), "Batch no GMSmyeloid-HD829")

    @validator("amount_needed", always=True)
    def get_amount_needed(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Amount needed (ng)")



import logging
import sys

import click

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_qc_output_artifacts

LOG = logging.getLogger(__name__)


@click.command()
@options.concentration_udf_option()
@options.amount_udf_option()
@options.volume_udf_option()
@options.subtract_volume_option()
@click.pass_context
def calculate_amount_ng(
    ctx: click.Context,
    amount_udf: str,
    volume_udf: str,
    concentration_udf: str,
    subtract_volume: str,
):
    """Calculates and auto-fills the quantities of DNA in sample from concentration and volume
    measurements. The volume is subtracted by either 0 or 3 in the calculations. This is
    because the lab uses 0 or 3 ul in the initial qc measurements."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        artifacts = get_qc_output_artifacts(lims=lims, process=process)
        missing_udfs_count = 0
        artifacts_with_missing_udf = []
        for artifact in artifacts:
            vol = artifact.udf.get(volume_udf)
            conc = artifact.udf.get(concentration_udf)
            if None in [conc, vol]:
                missing_udfs_count += 1
                artifacts_with_missing_udf.append(artifact.id)
                continue

            artifact.udf[amount_udf] = conc * (vol - int(subtract_volume))
            artifact.put()
        if missing_udfs_count:
            raise MissingUDFsError(
                f"Udf missing for {missing_udfs_count} artifact(s): "
                f"{','.join(artifacts_with_missing_udf)}. "
            )
        message = "Amounts have been calculated for all artifacts."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ProcessUDFs(BaseModel):
    """A-tailing and Adapter ligation (RNA) v1"""

    rna_prepp: str = Field(..., alias="RGT no: Illumina RNAprep")
    rna_index_anchors: str = Field(..., alias="Lot no: RNA index anchors")
    index: str = Field(..., alias="Lot no: Index")
    pcr_pre_lab: str = Field(..., alias="Thermal cycler (pre-lab)")
    water_pre_lab: str = Field(..., alias="Lot no: Nuclease-free water (pre-lab)")
    et_oh_pre_lab: str = Field(..., alias="Lot no: EtOH (pre-lab)")
    ampure_beads_pre_lab: str = Field(..., alias="Lot no: AMPure XP-beads (pre-lab)")
    pcr_amplification: str = Field(..., alias="Thermal cycler Amplification")
    cycles_pcr: int = Field(..., alias="Number of PCR cycles used")
    ampure_beads_post_lab: str = Field(..., alias="Lot no: AMPure XP-beads (post-lab)")
    et_oh_post_lab: str = Field(..., alias="Lot no: EtOH (post-lab)")
    water_post_lab: str = Field(..., alias="Lot no: Nuclease-free water (post-lab)")
    resuspension_buffer_post_lab: str = Field(..., alias="Lot no: Resuspension buffer (post-lab)")


class ArtifactUDFs(BaseModel):
    concentration: float = Field(..., alias="Concentration")
    size: Optional[int] = Field(None, alias="Size (bp)")


class ArnoldStep(
    BaseStep,
):
    artifact_udfs: ArtifactUDFs
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_a_tailning_and_adapter_ligation(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="A-tailing and Adapter ligation (RNA) v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="a_tailing_and_adapter_ligation",
        workflow="RNA"
    )



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class SampleArtifactUDFs(BaseModel):
    """Aggregate QC (RNA) v1"""

    sample_rin: float = Field(None, alias="RIN")
    sample_concentration: float = Field(None, alias="Concentration")
    sample_amount: float = Field(None, alias="Amount (ng)")


class SampleArtifactFields(BaseStep):
    artifact_udfs: SampleArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_sample_artifact_fields(
    lims: Lims, sample_id: str, prep_id: str
) -> Optional[SampleArtifactFields]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        optional_step=True,
    )
    sample_artifact_udfs = SampleArtifactUDFs(**analyte.artifact_udfs())
    if not sample_artifact_udfs.dict(exclude_none=True):
        return None

    return SampleArtifactFields(
        **analyte.base_fields(),
        artifact_udfs=sample_artifact_udfs,
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="reception_control",
        workflow="RNA"
    )



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFs(BaseModel):
    sample_normalization_method: Optional[str] = Field(None, alias="Method document")
    normalized_sample_concentration: Optional[float] = Field(
        None, alias="Final Concentration (ng/ul)"
    )
    lot_nr_dilution_buffer_sample_normalization: Optional[str] = Field(
        None, alias="Dilution buffer lot no"
    )


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_normalization_of_mictobial_samples(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="CG002 - Normalization of microbial samples",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="normalization",
        workflow="Microbial"
    )



import logging
import sys

import click
from genologics.entities import Artifact
from requests.exceptions import ConnectionError

from cg_lims.exceptions import LimsError, MissingCgFieldError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def set_artifact_rerun(
    target_amount_reads: float, guaranteed_fraction: float, reads_total: int, artifact: Artifact
):
    """Set Missing Reads, qc flag and Rerun"""
    target_amount = target_amount_reads / 1000000
    reads_min = target_amount * guaranteed_fraction
    reads_missing = reads_min - reads_total

    if reads_missing > 0:
        reads_to_sequence = target_amount - reads_total
        artifact.udf["Rerun"] = True
        artifact.qc_flag = "FAILED"
    else:
        reads_to_sequence = 0
        artifact.udf["Rerun"] = False
        artifact.qc_flag = "PASSED"

    for sample in artifact.samples:
        sample.udf["Reads missing (M)"] = reads_to_sequence
        sample.put()

    artifact.put()


def check_control(artifact: Artifact) -> bool:
    """Return True if all atrtifact samples are negative control samples."""
    return all(sample.udf.get("Control") == "negative" for sample in artifact.samples)


def find_reruns(artifacts: list, status_db) -> None:
    """
    Looking for artifacts to rerun.
    Negative control samples are never sent for rerun.
    A pool with any sample that is not a negative control will be sent for rerun if reads are missing."""
    failed_arts = 0
    for artifact in artifacts:
        if check_control(artifact):
            artifact.qc_flag = "PASSED"
            artifact.put()
            continue
        sample = artifact.samples[0]
        reads_total = sample.udf.get("Total Reads (M)")
        app_tag = sample.udf.get("Sequencing Analysis")

        if None in (app_tag, reads_total):
            failed_arts += 1
            continue

        try:
            target_amount_reads = status_db.apptag(tag_name=app_tag, key="target_reads")
            guaranteed_fraction = 0.01 * status_db.apptag(
                tag_name=app_tag, key="percent_reads_guaranteed"
            )
        except ConnectionError:
            raise LimsError(message="Could not communicate with cg server")
        except:
            raise MissingCgFieldError(f"Could not find application tag: {app_tag} in database.")

        set_artifact_rerun(target_amount_reads, guaranteed_fraction, reads_total, artifact)

    if failed_arts:
        raise MissingUDFsError(
            message=f"Could not update {failed_arts} artifacts due to missing UDFs"
        )


@click.command()
@click.pass_context
def get_missing_reads(ctx):
    """Script to calculate missing reads and decide on reruns.
    Negative control samples are never sent for rerun.
    A pool with any sample that is not a negative control will be sent for rerun if reads are missing."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]
    status_db = ctx.obj["status_db"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        find_reruns(artifacts=artifacts, status_db=status_db)
        message = "Missing Reads and Rerun info have been set on all artifacts"
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from typing import List
from genologics.entities import Artifact
from genologics.lims import Lims

from cg_lims.get.artifacts import get_latest_artifact
from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.set.udfs import copy_udf_process_to_artifact

import logging
import sys
import click

LOG = logging.getLogger(__name__)


def copy_udf_to_artifacts(
    artifacts: List[Artifact],
    process_types: List[str],
    lims: Lims,
    artifact_udf: str,
    process_udf: str,
):
    """Looping over all artifacts. Getting the latest artifact to copy from. Copying."""

    failed_artifacts = 0
    for destination_artifact in artifacts:
        try:
            sample = destination_artifact.samples[0]
            source_artifact: Artifact = get_latest_artifact(
                lims=lims, sample_id=sample.id, process_types=process_types
            )
            source_process = source_artifact.parent_process
            copy_udf_process_to_artifact(
                artifact_udf=artifact_udf,
                process_udf=process_udf,
                artifact=destination_artifact,
                process=source_process,
            )
        except:
            failed_artifacts += 1
    if failed_artifacts:
        raise MissingUDFsError(
            message=f"Failed to set artifact udfs on {failed_artifacts} artifacts. See log for details"
        )


@click.command()
@options.process_types()
@options.artifact_udf()
@options.process_udf()
@options.measurement(help="Udfs will be set on measurements.")
@options.input(help="Udfs will be set on input artifacts.")
@click.pass_context
def process_to_artifact(
    ctx,
    artifact_udf: str,
    process_udf: str,
    measurement: bool,
    input: bool,
    process_types: List[str],
):
    """Script to copy udf to destination artifacts in the current step,
    from process udf defined by process_types and process_udf

    The input and measurement arguments define what kind of artifacts from the current step that
    should be the destination artifacts.
    """

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        artifacts = get_artifacts(process=process, input=input, measurement=measurement)
        copy_udf_to_artifacts(
            artifacts=artifacts,
            artifact_udf=artifact_udf,
            process_udf=process_udf,
            process_types=process_types,
            lims=lims,
        )
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



import logging
import sys
from typing import List

import click

from cg_lims import options
from cg_lims.get.artifacts import get_latest_artifact, get_qc_output_artifacts

LOG = logging.getLogger(__name__)


@click.command()
@options.process_types(help="Get buffer from this process type(s)")
@click.pass_context
def get_volumes_from_buffer(ctx, process_types: List[str]) -> None:
    """Getting Volume Elution from previous step of type defined by process_types.
    If volume found, setting the value Volume udf on artifact of current step.
    As part of the sample is left in the buffer excgange step,
    the value is subtracted by 5 to get the actual volume that is left."""

    LOG.info(f"Running {ctx.command_path} with params {ctx.params}")

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    failed_count = 0
    updated_count = 0
    artifacts = get_qc_output_artifacts(lims, process)
    for artifact in artifacts:
        try:
            buffer_artifact = get_latest_artifact(
                lims=lims, sample_id=artifact.samples[0].id, process_types=list(process_types)
            )
        except:
            continue
        if buffer_artifact.udf.get("Volume Elution (ul)"):
            volume_buffer = float(buffer_artifact.udf.get("Volume Elution (ul)")) - 5.0
            if volume_buffer < 0:
                failed_count += 1
                continue
            updated_count += 1
            artifact.udf["Volume (ul)"] = volume_buffer
            artifact.put()

    message = f"Updated {updated_count} samples with volume from Buffer step."
    LOG.info(message)
    click.echo(message)

    if failed_count > 0:
        message = f"Failed to update {failed_count} artifacts - buffer volume lower than 0!"
        LOG.error(message)
        sys.exit(message)



import logging
import sys
import click
import numpy as np

from typing import List

from genologics.entities import Artifact

from cg_lims import options
from cg_lims.get.artifacts import get_artifacts
from cg_lims.exceptions import LimsError, MissingUDFsError

LOG = logging.getLogger(__name__)


def calculate_average_size(all_artifacts: List[Artifact]) -> float:
    """Calculate average size (bp) from given list of artifacts.
    Returns the average and number of missing UDF:s."""

    artifacts = [artifact for artifact in all_artifacts if artifact.name[0:3] != "NTC"]
    size_list = [
        artifact.udf["Size (bp)"]
        for artifact in artifacts
        if artifact.udf.get("Size (bp)") is not None
    ]

    if not size_list:
        raise MissingUDFsError("Udf missing for 'Size (bp)' sample(s).")
    return float(np.mean(size_list))


def set_average_and_qc(
    average: float,
    lower_threshold: str,
    upper_threshold: str,
    all_artifacts: List[Artifact],
) -> None:
    """Sets the given average to a list of artifacts.
    Also assesses QC flags according to defined thresholds."""
    qc_flag = "PASSED"
    if lower_threshold is not None and int(average) < int(lower_threshold):
        qc_flag = "FAILED"
    elif upper_threshold and int(average) > int(upper_threshold):
        qc_flag = "FAILED"

    for artifact in all_artifacts:
        artifact.udf["Average Size (bp)"] = str(average)
        if artifact.qc_flag != "FAILED":
            artifact.qc_flag = qc_flag
        artifact.put()


@click.command()
@options.lower_threshold()
@options.upper_threshold()
@click.pass_context
def calculate_average_size_and_set_qc(ctx, lower_threshold: str, upper_threshold: str) -> None:
    """Script to calculate and apply Average Size (bp) to a set of samples
    given a subset of their real size lengths. NTC samples are ignored in the calculations."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]

    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=True)
        average: float = calculate_average_size(all_artifacts=artifacts)
        set_average_and_qc(
            average=average,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
            all_artifacts=artifacts,
        )
        message = "Average Size (bp) have been calculated and set for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python
from cg_lims.exceptions import LimsError, MissingUDFsError

from cg_lims.get.artifacts import get_artifacts, get_qc_output_artifacts
from cg_lims.get.udfs import get_udf_type
from cg_lims import options

import logging
import sys

import click

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def udf_copy_sample_to_artifact(artifacts: list, sample_udf: str, art_udf: str, lims) -> None:
    """Function to copy sample udf to artifact.

    For each artifact in the artifacts list, picking the first sample in the art.samples list.
    copying the samp_udf from the sample to the art_udf on the artifact.
    Warns if sample_udf is missing on a sample.
    Logs if a the sample list of a artifact contains more than one sample.


    Arguments:
        artifacts: list of artifacts to copy from
        art_udf: artifact udf to get
        sample_udf: sample udf to set"""
    failed_udfs = 0
    passed_udfs = 0
    udf_type = get_udf_type(lims=lims, udf_name=art_udf, attach_to_name=artifacts[0].output_type)
    for art in artifacts:
        if len(art.samples) > 1:
            LOG.info("More than one sample per artifact, picking the first one")
        sample = art.samples[0]
        udf = sample.udf.get(sample_udf)
        if not isinstance(udf, udf_type):
            try:
                udf = udf_type(udf)
            except:
                failed_udfs += 1
                LOG.error(
                    f"Udf: {sample_udf} missing on sample {sample.id} or it could not be converted to type {udf_type}"
                )
                continue

        art.udf[art_udf] = udf
        art.put()
        LOG.info(
            f"copied udf {sample_udf} from sample {sample.id} to artifact: {art.id}, udf: {art_udf}"
        )
        passed_udfs += 1

    if failed_udfs:
        raise MissingUDFsError(
            message=f"The udf '{sample_udf}' is missing for {failed_udfs} samples. Udfs were set on {passed_udfs} artifacts."
        )


@click.command()
@options.sample_udf(help="Sample udf to set.")
@options.artifact_udf(help="Artifact udf to get.")
@options.measurement(help="Udfs will be set on measurements.")
@options.input(help="Udfs will be set on input artifacts.")
@click.pass_context
def sample_to_artifact(ctx, sample_udf, artifact_udf, measurement, input):
    """Script to copy artifact udf to sample udf"""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        artifacts = get_artifacts(process=process, input=input, measurement=measurement)
        udf_copy_sample_to_artifact(artifacts, sample_udf, artifact_udf, lims)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ArtifactUDF(BaseModel):
    sample_amount_needed: float = Field(..., alias="Amount needed (ng)")


class ArnoldStep(BaseStep):
    artifact_udfs: ArtifactUDF

    class Config:
        allow_population_by_field_name = True


def get_aliquot_samples_for_covaris(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Aliquot Samples for Covaris",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        artifact_udfs=ArtifactUDF(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="aliquot_samples_for_covaris",
        workflow="WGS",
    )



from typing import List, Optional

from genologics.entities import Artifact, Process, Sample
from genologics.lims import Lims

from cg_lims.exceptions import FileError, MissingArtifactError
from datetime import datetime
import logging

LOG = logging.getLogger(__name__)


def get_sample_artifact(lims: Lims, sample: Sample) -> Artifact:
    """Returning the initial artifact related to a sample.
    Assuming first artifact is allways named sample.id + 'PA1."""
    sample_artifact_id = f"{sample.id}PA1"
    artifact = Artifact(lims, id=sample_artifact_id)
    try:
        artifact.get()
    except:
        message = f"no artifact with id {sample_artifact_id}"
        LOG.error(message)
        raise MissingArtifactError(message=message)
    return artifact


def get_qc_output_artifacts(lims: Lims, process: Process) -> List[Artifact]:
    """Get output 'artifacts' (messuements) of a qc process"""

    input_output_maps = process.input_output_maps
    artifact_ids = [
        io[1]["limsid"] for io in input_output_maps if io[1]["output-generation-type"] == "PerInput"
    ]
    return [Artifact(lims, id=id) for id in artifact_ids if id is not None]


def get_artifacts(
    process: Process, input: Optional[bool] = False, measurement: Optional[bool] = False
) -> List[Artifact]:
    """If inputs is True, returning all input analytes of the process,
    otherwise returning all output analytes of the process"""

    if measurement:
        return get_qc_output_artifacts(lims=process.lims, process=process)
    elif input:
        return process.all_inputs(unique=True)
    else:
        return [a for a in process.all_outputs(unique=True) if a.type == "Analyte"]


def get_artifact_by_name(process: Process, name: str) -> Artifact:
    """Searches for output artifacts with name <file_name>
    Raises error if more than one artifact with same name!"""

    artifacts = list(filter(lambda a: a.name in [name], process.all_outputs()))

    if len(artifacts) > 1:
        message = f"more than one output artifact named {name}"
        LOG.error(message)
        raise FileError(message=message)

    return artifacts[0]


def filter_artifacts(artifacts: List[Artifact], udf: str, value) -> List[Artifact]:
    """Returning a list of only artifacts with udf==value"""

    return [a for a in artifacts if a.udf.get(udf) == value]


def get_latest_artifact(
    lims: Lims, sample_id: str, process_types: Optional[List[str]], sample_artifact: bool = False
) -> Artifact:
    """Getting the most recently generated artifact by process_types and sample_id.

    Searching for all artifacts (Analytes) associated with <sample_id> that
    were produced by <process_types>.

    Returning the artifact with latest parent_process.date_run.
    If there are many such artifacts only one will be returned."""

    if sample_artifact and not process_types:
        return get_sample_artifact(lims=lims, sample=Sample(lims, sample_id))

    lims_artifacts = lims.get_artifacts(
        samplelimsid=sample_id,
        type="Analyte",
        process_type=process_types,
    )

    if sample_artifact and not lims_artifacts:
        return get_sample_artifact(lims=lims, sample=Sample(lims, sample_id))

    if not lims_artifacts:
        message = f"Could not find a artifact with sample: {sample_id} generated by process: {' ,'.join(process_types)}"
        LOG.error(message)
        raise MissingArtifactError(message=message)

    artifacts = []
    for artifact in lims_artifacts:
        date = artifact.parent_process.date_run or datetime.today().strftime("%Y-%m-%d")
        artifacts.append((date, artifact.id, artifact))

    artifacts.sort()
    date, id, latest_art = artifacts[-1]

    return latest_art



from typing import List, Optional, Tuple, Iterator
from genologics.entities import Artifact
from genologics.lims import Lims

from cg_lims.get.artifacts import get_latest_artifact
from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError, ArgumentError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.set.udfs import copy_artifact_to_artifact

import logging
import sys
import click

LOG = logging.getLogger(__name__)


def copy_udfs_to_all_artifacts(
    artifacts: List[Artifact],
    process_types: List[str],
    lims: Lims,
    udfs: List[Tuple[str, str]],
    sample_artifact: bool = False,
    qc_flag: bool = False,
):
    """Looping over all artifacts. Getting the latest artifact to copy from. Copying."""

    failed_artifacts = 0
    for destination_artifact in artifacts:
        try:
            sample = destination_artifact.samples[0]
            source_artifact = get_latest_artifact(
                lims=lims,
                sample_id=sample.id,
                process_types=process_types,
                sample_artifact=sample_artifact,
            )
            copy_artifact_to_artifact(
                destination_artifact=destination_artifact,
                source_artifact=source_artifact,
                artifact_udfs=udfs,
                qc_flag=qc_flag,
            )
        except:
            failed_artifacts += 1
    if failed_artifacts:
        raise MissingUDFsError(
            message=f"Failed to set artifact udfs on {failed_artifacts} artifacts. See log for details"
        )


@click.command()
@options.process_types()
@options.sample_artifact()
@options.artifact_udfs()
@options.source_artifact_udfs()
@options.measurement(help="Udfs will be set on measurements.")
@options.input(help="Udfs will be set on input artifacts.")
@options.qc_flag()
@click.pass_context
def artifact_to_artifact(
    ctx,
    artifact_udfs: List[str],
    source_artifact_udfs: Optional[List[str]],
    measurement: bool,
    input: bool,
    process_types: List[str],
    sample_artifact: bool,
    qc_flag: bool,
):
    """Script to copy udfs to destination artifacts in the current step, from source artifacts
    generated by steps defined by process_types, or from the original sample artifact if the sample-artifact
     flagg is true.

    The input and measurement arguments define what kind of artifacts from the current step that
    should be the destination artifacts.
    """

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    if not source_artifact_udfs:
        source_artifact_udfs = artifact_udfs
    if len(source_artifact_udfs) != len(artifact_udfs):
        raise ArgumentError(
            "The number of artifact-udfs and source-artifact-udfs must be the same."
        )

    udf_pairs = zip(source_artifact_udfs, artifact_udfs)
    try:
        artifacts = get_artifacts(process=process, input=input, measurement=measurement)
        copy_udfs_to_all_artifacts(
            artifacts=artifacts,
            udfs=list(udf_pairs),
            process_types=process_types,
            lims=lims,
            sample_artifact=sample_artifact,
            qc_flag=qc_flag,
        )
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



import mock
import pytest

from cg_lims.EPPs.udf.calculate.calculate_beads import (
    ELUTION_VOLUME,
    ELUTION_VOLUME_FACTOR,
    SAMPLE_VOLUME_BOUNDARY,
    calculate_beads_volume,
    calculate_elution_volume,
    calculate_volumes,
    calculate_water_volume,
)
from cg_lims.exceptions import MissingUDFsError
from genologics.entities import Artifact
from genologics.lims import Lims
from tests.conftest import server

OFFSET = 20


@pytest.mark.parametrize(
    "sample_volume, return_value",
    [
        (SAMPLE_VOLUME_BOUNDARY - OFFSET, ELUTION_VOLUME),
        (SAMPLE_VOLUME_BOUNDARY, ELUTION_VOLUME_FACTOR * SAMPLE_VOLUME_BOUNDARY),
        (
            SAMPLE_VOLUME_BOUNDARY + OFFSET,
            ELUTION_VOLUME_FACTOR * (SAMPLE_VOLUME_BOUNDARY + OFFSET),
        ),
    ],
)
def test_calculate_elution_volume(sample_volume: float, return_value: float):
    # GIVEN a sample volume less than, equal to, and greater than SAMPLE_VOLUME_BOUNDARY

    # WHEN calculating the elution volume
    result = calculate_elution_volume(sample_volume)

    # THEN the elution volume should be calculated correctly
    assert result == return_value
    assert isinstance(result, float)


@pytest.mark.parametrize(
    "sample_volume, expected_return_value",
    [
        (SAMPLE_VOLUME_BOUNDARY - OFFSET, OFFSET),
        (SAMPLE_VOLUME_BOUNDARY, 0.0),
        (SAMPLE_VOLUME_BOUNDARY + OFFSET, 0.0),
    ],
)
def test_calculate_water_volume(sample_volume: float, expected_return_value: float):
    # GIVEN a sample volume less than, equal to, and greater than SAMPLE_VOLUME_BOUNDARY

    # WHEN calculating the water volume
    result = calculate_water_volume(sample_volume)

    # THEN the water volume should be calculated correctly
    assert result == expected_return_value
    assert isinstance(result, float)


@pytest.mark.parametrize(
    "sample_volume, water_volume, expected_return_value",
    [
        (50.0, 50.0, 200.0),
        (0.0, 50.0, 100.0),
        (50.0, 0.0, 100.0),
    ],
)
def test_calculate_beads_volume(
    sample_volume: float, water_volume: float, expected_return_value: float
):
    # GIVEN a sample_volume and a water_volume

    # WHEN calculating the beads volume
    result = calculate_beads_volume(
        sample_volume=sample_volume, h2o_volume=water_volume
    )

    # THEN calculate beads volume should be twice the sum of the sample volume and water volume
    assert result == expected_return_value
    assert isinstance(result, float)


@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_elution_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_water_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_beads_volume")
@pytest.mark.parametrize(
    "sample_volume, elution_volume, water_volume, beads_volume",
    [(20.0, 40.0, 30.0, 100.0)],
)
def test_calculate_volumes_single_artifact(
    mock_beads_volume,
    mock_water_volume,
    mock_elution_volume,
    sample_volume: float,
    elution_volume: float,
    water_volume: float,
    beads_volume: float,
    artifact_1: Artifact,
):
    # GIVEN an artifact with a udf "Sample Volume (ul)"
    server("flat_tests")
    artifact_1.udf["Sample Volume (ul)"] = sample_volume
    artifacts = [artifact_1]
    mock_elution_volume.return_value = elution_volume
    mock_water_volume.return_value = water_volume
    mock_beads_volume.return_value = beads_volume

    # WHEN calculating the volumes
    calculate_volumes(artifacts=artifacts)

    # THEN the calculated volumes for elution, water and beads are correct
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume
    assert artifact_1.udf.get("Volume Elution (ul)") == elution_volume
    assert artifact_1.udf.get("Volume H2O (ul)") == water_volume
    assert artifact_1.udf.get("Volume Beads (ul)") == beads_volume


@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_elution_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_water_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_beads_volume")
@pytest.mark.parametrize(
    "sample_volume_1, elution_volume_1, water_volume_1, beads_volume_1,"
    "sample_volume_2, elution_volume_2, water_volume_2, beads_volume_2",
    [(20.0, 40.0, 30.0, 100.0, 60.0, 48.0, 0.0, 120.0)],
)
def test_calculate_volumes_multiple_artifacts(
    mock_beads_volume,
    mock_water_volume,
    mock_elution_volume,
    sample_volume_1: float,
    elution_volume_1: float,
    water_volume_1: float,
    beads_volume_1: float,
    sample_volume_2: float,
    elution_volume_2: float,
    water_volume_2: float,
    beads_volume_2: float,
    artifact_1: Artifact,
    artifact_2: Artifact,
):
    # GIVEN two artifact with a udf "Sample Volume (ul)"
    server("flat_tests")
    artifact_1.udf["Sample Volume (ul)"] = sample_volume_1
    artifact_2.udf["Sample Volume (ul)"] = sample_volume_2

    artifacts = [artifact_1, artifact_2]

    mock_elution_volume.side_effect = [elution_volume_1, elution_volume_2]
    mock_water_volume.side_effect = [water_volume_1, water_volume_2]
    mock_beads_volume.side_effect = [beads_volume_1, beads_volume_2]

    # WHEN calculating the volumes
    calculate_volumes(artifacts=artifacts)

    # THEN the calculated volumes for elution, water and beads should be assigned correctly
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume_1
    assert artifact_1.udf.get("Volume Elution (ul)") == elution_volume_1
    assert artifact_1.udf.get("Volume H2O (ul)") == water_volume_1
    assert artifact_1.udf.get("Volume Beads (ul)") == beads_volume_1
    assert artifact_2.udf.get("Sample Volume (ul)") == sample_volume_2
    assert artifact_2.udf.get("Volume Elution (ul)") == elution_volume_2
    assert artifact_2.udf.get("Volume H2O (ul)") == water_volume_2
    assert artifact_2.udf.get("Volume Beads (ul)") == beads_volume_2


def test_calculate_volumes_single_artifact_missing_sample_volume_udf(
    artifact_1: Artifact,
):
    # GIVEN an artifact with no udf "Sample Volume (ul)"
    server("flat_tests")
    assert artifact_1.udf.get("Sample Volume (ul)") is None
    artifacts = [artifact_1]

    # WHEN calculating the volumes
    # THEN an exception should be raised
    with pytest.raises(MissingUDFsError) as error_message:
        calculate_volumes(artifacts=artifacts)
    assert f"missing for 1 out of 1 samples" in error_message.value.message


@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_elution_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_water_volume")
@mock.patch("cg_lims.EPPs.udf.calculate.calculate_beads.calculate_beads_volume")
@pytest.mark.parametrize(
    "sample_volume, elution_volume, water_volume, beads_volume",
    [(20.0, 40.0, 30.0, 100.0)],
)
def test_calculate_volumes_multiple_artifacts_missing_sample_volume_udf(
    mock_beads_volume,
    mock_water_volume,
    mock_elution_volume,
    sample_volume: float,
    elution_volume: float,
    water_volume: float,
    beads_volume: float,
    lims: Lims,
    artifact_1: Artifact,
    artifact_2: Artifact,
):
    # GIVEN one out of two artifacts has no udf "Sample Volume (ul)"
    server("flat_tests")
    artifact_1.udf["Sample Volume (ul)"] = sample_volume
    assert artifact_1.udf.get("Sample Volume (ul)") is not None
    assert artifact_2.udf.get("Sample Volume (ul)") is None
    artifacts = [artifact_1, artifact_2]

    mock_elution_volume.return_value = elution_volume
    mock_water_volume.return_value = water_volume
    mock_beads_volume.return_value = beads_volume

    # WHEN calculating the volumes
    with pytest.raises(MissingUDFsError) as error_message:
        calculate_volumes(artifacts=artifacts)

    # THEN an exception should be raised
    assert f"missing for 1 out of 2 samples" in error_message.value.message

    # AND the artifact with the udf "Sample Volume (ul) should be processed correctly
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume
    assert artifact_1.udf.get("Sample Volume (ul)") == sample_volume
    assert artifact_1.udf.get("Volume Elution (ul)") == elution_volume
    assert artifact_1.udf.get("Volume H2O (ul)") == water_volume



import logging
from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.prep.base_step import BaseStep
from cg_lims.objects import BaseAnalyte

LOG = logging.getLogger(__name__)


class ProcessUDFs(BaseModel):
    lot_nr_beads_library_prep: str = Field(..., alias="Lot nr: Beads")
    lot_nr_etoh_library_prep: str = Field(..., alias="Lot nr: EtOH")
    lot_nr_h2o_library_prep: str = Field(..., alias="Lot nr: H2O")


class ArtifactUDFs(BaseModel):
    finished_library_concentration: float = Field(..., alias="Concentration")
    finished_library_concentration_nm: float = Field(..., alias="Concentration (nM)")
    finished_library_size: Optional[int] = Field(None, alias="Size (bp)")
    finished_library_average_size: float = Field(..., alias="Average Size (bp)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_post_bead_pcr_purification(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Post-PCR bead purification v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="post_pcr_bead_purification",
        workflow="Microbial"
    )



from typing import Optional

from genologics.lims import Lims
from pydantic import Field, BaseModel
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ArtifactUDFs(BaseModel):
    bait_set: Optional[str] = Field(None, alias="Bait Set")
    capture_kit: Optional[str] = Field(None, alias="Capture kit lot nr.")


class ProcessUDFs(BaseModel):
    lot_nr_blockers: str = Field(..., alias="Blockers")
    lot_nr_hybridization_kit: str = Field(..., alias="TWIST Hybridization kit")
    pcr_instrument_hybridization: str = Field(..., alias="Thermal cycler")
    hybridization_method: str = Field(..., alias="Method document")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_hybridize_library_twist(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Hybridize Library TWIST v2",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="hybridize_library",
        workflow="TWIST",
    )



from genologics.entities import Process

from cg_lims.EPPs.arnold.prep import build_step_documents
from cg_lims.models.arnold.prep.base_step import BaseStep

from tests.conftest import (
    server,
)


WGS_STEP_TYPES = {
    "reception_control",
    "end_repair",
    "fragment_dna",
    "aliquot_samples_for_covaris",
}


def test_wgs_prep(lims):
    # GIVEN: A lims with a process: "24-240276" (Aggregate QC (Library Validation)).
    # Where the samples in the process has gone through the whole wgs prep workflow:

    server("wgs_prep")
    process = Process(lims=lims, id="24-240276")

    # WHEN running build_step_documents
    step_documents = build_step_documents(lims=lims, process=process, prep_type="wgs")
    # THEN assert BaseStep documents are created and all step types in the microbial workflow are represented
    for document in step_documents:
        assert isinstance(document, BaseStep)
    assert {document.step_type for document in step_documents} == WGS_STEP_TYPES



from typing import Optional

from genologics.entities import Artifact, Process
from pydantic import Field, validator

from cg_lims.models.api.master_steps.base_step import (
    get_artifact_udf,
    BaseStep,
    get_process_udf,
)
from cg_lims.get.artifacts import get_latest_artifact


class HybridizeLibraryTWIST(BaseStep):
    "Hybridize Library TWIST v2"
    artifact: Optional[Artifact]
    process: Optional[Process]
    bait_set: Optional[str] = Field(None, alias="Bait Set")
    capture_kit: Optional[str] = Field(None, alias="Capture kit lot nr.")
    container_name: Optional[str] = Field(None, alias="Container Name")
    enrichment_kit: Optional[str] = Field(None, alias="Twist enrichment kit")
    blockers: Optional[str] = Field(None, alias="Blockers")
    hybridization_kit: Optional[str] = Field(None, alias="TWIST Hybridization kit")
    thermal_cycler: Optional[str] = Field(None, alias="Thermal cycler")
    method_document: Optional[str] = Field(None, alias="Method document")
    document_version: Optional[str] = Field(None, alias="Document version")

    @validator("artifact", always=True)
    def get_artifact(cls, v, values):
        try:
            return get_latest_artifact(
                lims=values.get("lims"),
                sample_id=values.get("sample_id"),
                process_types=["Hybridize Library TWIST v2"],
            )
        except:
            return None

    @validator("process", always=True)
    def get_process(cls, v, values):
        if values.get("artifact"):
            return values.get("artifact").parent_process
        return None

    @validator("bait_set", always=True)
    def get_bait_set(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Bait Set")

    @validator("capture_kit", always=True)
    def get_capture_kit(cls, v, values):
        return get_artifact_udf(values.get("artifact"), "Capture kit lot nr.")

    @validator("container_name", always=True)
    def get_container_name(cls, v, values):
        artifact = values.get("artifact")
        if artifact:
            return artifact.container.name
        return None

    @validator("enrichment_kit", always=True)
    def get_enrichment_kit(cls, v, values):
        return get_process_udf(values.get("process"), "Twist enrichment kit")

    @validator("blockers", always=True)
    def get_blockers(cls, v, values):
        return get_process_udf(values.get("process"), "Blockers")

    @validator("hybridization_kit", always=True)
    def get_hybridization_kit(cls, v, values):
        return get_process_udf(values.get("process"), "TWIST Hybridization kit")

    @validator("thermal_cycler", always=True)
    def get_thermal_cycler(cls, v, values):
        return get_process_udf(values.get("process"), "Thermal cycler")

    @validator("method_document", always=True)
    def get_method_document(cls, v, values):
        return get_process_udf(values.get("process"), "Method document")

    @validator("document_version", always=True)
    def get_document_version(cls, v, values):
        return get_process_udf(values.get("process"), "Document version")



import click
from genologics.lims import Lims

from cg_lims import options
import logging

LOG = logging.getLogger(__name__)


def search_automations(lims: Lims, search_by: str) -> None:
    """Script to search for Automations.
    The argument search_by could be a script name, eg: bcl2fastq.py,
    or a script argument, eg: CG002 - Aggregate QC (Library Validation).
    """
    LOG.info("AUTOMATION: Searching for '%s'. \n" % search_by)
    automations = lims.get_automations()
    for automation in automations:
        bash_string = automation.string
        if bash_string.find(search_by) == -1:
            continue

        process_ids = [
            (process_type.id, process_type.name) for process_type in automation.process_types
        ]
        LOG.info("Button: %s" % automation.name)
        LOG.info("Bash string: %s" % bash_string)
        LOG.info("Processes: %s \n" % process_ids)


@click.command()
@options.automation_string()
@click.pass_context
def check_config(ctx, automation_string: str):
    """Script to search for Automations.
    The argument search_by could be a script name, eg: bcl2fastq.py,
    or a script argument, eg: CG002 - Aggregate QC (Library Validation).
    """

    lims = ctx.obj["lims"]
    search_automations(search_by=automation_string, lims=lims)



#!/usr/bin/env python
import logging
import pathlib

import click
from cg_lims import options

# commands
from .check_config import check_config
from .prepare_fixture import make_fixure


@click.group(invoke_without_command=True)
@options.log()
@click.pass_context
def scripts(ctx, log: str):
    """Main entry point to run scripts"""

    log_path = pathlib.Path(log)
    logging.basicConfig(filename=str(log_path.absolute()), filemode="a", level=logging.INFO)


scripts.add_command(check_config)
scripts.add_command(make_fixure)



from typing import Optional
from genologics.lims import Lims
from pydantic import Field, BaseModel
from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.prep.base_step import BaseStep


class ProcessUDFs(BaseModel):
    amplify_captured_library_method: str = Field(..., alias="Method Document")
    lot_nr_xgen_primer_amplify_captured_library: str = Field(..., alias="xGen Library Amp primer")
    lot_nr_amplification_kit_amplify_captured_library: str = Field(
        ..., alias="Kapa HiFi HotStart ReadyMix"
    )
    nr_pcr_cycles_amplify_captured_library: int = Field(..., alias="Nr of PCR cycles")


class ArnoldStep(
    BaseStep,
):
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_amplify_captured_library_udfs(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Amplify Captured Libraries TWIST v2",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="amplify_captured_library",
        workflow="TWIST",
    )
